(defpackage :cl-gserver.actor-mp-test
  (:use :cl :trivia :iterate :fiveam :act)
  (:export #:run!
           #:all-tests
           #:nil))

(in-package :cl-gserver.actor-mp-test)

(def-suite actor-mp-tests
  :description "actor mp tests"
  :in cl-gserver.tests:test-suite)

(in-suite actor-mp-tests)

(log:config :warn)

(defvar *behavior* (lambda (self message current-state)
                        (declare (ignore self))
                        (match message
                          (:add
                           (let ((new-state (1+ current-state)))
                             (cons new-state new-state)))
                          (:sub
                           (let ((new-state (1- current-state)))
                             (cons new-state new-state)))
                          (:get (cons current-state current-state)))))

(def-fixture mp-setup (queue-size)
  (setf lparallel:*kernel* (lparallel:make-kernel 8))

  (defclass counter-actor (actor) ())
  
  (format t "Running non-system tests...~%")
  (let* ((cut (make-instance 'counter-actor
                             :name "counter-actor"
                             :state 0
                             :receive *behavior*))
         (max-loop 10000)
         (per-thread (/ max-loop 8)))
    (setf (act-cell:msgbox cut) (make-instance 'mesgb:message-box/bt :max-queue-size queue-size))
    (&body)
    (ask-s cut :stop))
  (format t "Running non-system tests...done~%")
  (format t "Running system tests...~%")
  (let* ((system (asys:make-actor-system :shared-dispatcher-workers 4))
         (cut (ac:actor-of system (lambda ()
                                    (make-instance 'counter-actor :state 0
                                                                  :receive *behavior*))))
         (max-loop 10000)
         (per-thread (/ max-loop 8)))
    (unwind-protect
         (&body)
    (ask-s cut :stop)
    (ac:shutdown system)))
  (format t "Running system tests...~%")
  
  (lparallel:end-kernel))


(test counter-mp-unbounded
  "Counter server - multi processors - unbounded queue"

  (with-fixture mp-setup (nil)
    ;; add
    (map nil #'lparallel:force
         (mapcar (lambda (x)
                   (declare (ignore x))
                   (lparallel:future
                     (dotimes (n (1+ per-thread))
                       (ask-s cut :add))
                     (dotimes (n per-thread)
                       (ask-s cut :sub))))
                 (loop repeat 8 collect "n")))
    (is (= 8 (ask-s cut :get)))))

(test counter-mp-bounded
  "Counter server - multi processors - bounded queue"

  (with-fixture mp-setup (100)
    ;; add
    (map nil #'lparallel:force
         (mapcar (lambda (x)
                   (declare (ignore x))
                   (lparallel:future
                     (dotimes (n (1+ per-thread))
                       (ask-s cut :add))
                     (dotimes (n per-thread)
                       (ask-s cut :sub))))
                 (loop repeat 8 collect "n")))
    (is (= 8 (ask-s cut :get)))))

(defun run-tests ()
  (time (run! 'counter-mp-unbounded))
  (time (run! 'counter-mp-bounded)))
